# -----------------------------------------------------------
# Copyright (c) 2018 Lenovo AI Platform. All Rights Reserved.
# -----------------------------------------------------------

'''
Example: 
python kvmem_inference.py --input_file=/home/xh/data/kvmem_test.txt --checkpoint_dir='/home/xh/data/kvmem' --output_dir=/tmp
Test text:(question must end with two \t)
1 Mary moved to the bathroom.
2 John went to the hallway.
3 Daniel went back to the hallway.
4 Sandra moved to the garden.
5 John moved to the office.
6 Sandra journeyed to the bathroom.
7 Where is Sandra?
8 Mary moved to the hallway.
9 Daniel travelled to the office.
10 John went back to the garden.
11 John moved to the bedroom.
12 Daniel moved to the garden.
13 Where is Daniel?
1 Daniel journeyed to the kitchen.
2 Daniel journeyed to the bedroom.
3 Where is Daniel?
4 Sandra moved to the garden.
5 Sandra went back to the kitchen.
6 Where is Sandra?
'''

from __future__ import division

import math
import time
import os
import sys
import numpy as np
from itertools import chain
from datetime import datetime

import tensorflow as tf
import tensorflow.python.platform
from tensorflow.python.eager import context
from tensorflow.python.platform import gfile

parentpath = os.path.dirname(os.path.abspath(__file__))
path_list = [parentpath[:-19] + 'lib',
             parentpath[:-19] + 'engine',
             parentpath[:-19] + 'applications',
             parentpath[:-19] + 'applications/' + 'kvmem']
sys.path += path_list
print sys.path
import get_kv_data, get_kv_loss

slim = tf.contrib.slim
FLAGS = tf.app.flags.FLAGS
tf.logging.set_verbosity(tf.logging.INFO)

VOCAB = []
VOCAB_TXT = FLAGS.checkpoint_dir + '/vocab.txt'
SIZE_TXT = FLAGS.checkpoint_dir + '/size.txt'

def pre_check():
    '''pre check all flags to make sure following work can be executed.
    '''
    assert (FLAGS.memory_size and FLAGS.embedding_size and FLAGS.hops
            and FLAGS.l2_lambda and FLAGS.feature_size and FLAGS.batch_size)

    if not FLAGS.checkpoint_dir:
      raise ValueError("Checkpoint dir must be provided to restore model.")
    if not FLAGS.input_file:
      raise ValueError("Target inference file must be provided.")
    if not FLAGS.output_dir:
      raise ValueError("Output directory for answers must be provided.")
    if FLAGS.output_dir[-1] != '/':
      FLAGS.output_dir = FLAGS.output_dir + '/'
    if not os.path.exists(VOCAB_TXT) or not os.path.exists(SIZE_TXT):
      raise ValueError("Trained vocab.txt and size.txt not found in ",
                       FLAGS.checkpoint_dir)

def load(saver, sess, ckpt_path):
    '''Load trained weights.
    
    Args:
      saver: TensorFlow saver object.
      sess: TensorFlow session.
      ckpt_path: path to checkpoint file with parameters.
    '''
    saver.restore(sess, ckpt_path)
    print("Restored model parameters from {}".format(ckpt_path))

def load_infer_query(filename, only_supporting=False):
    """Load inference query from inference file 

    Args:
        filename: inference file name (full path)

    Return:
        parsed story, query and answer.
    """
    res = []
    with open(filename) as f:
      data = get_kv_data.parse_stories(f.readlines(),
                                          only_supporting=only_supporting)
      for story, query, answer in data:
        if not answer or answer[0] == '':
          answer = []
        res.append((story, query, answer))
      return res

def create_model():
    # Get story and query
    with open(VOCAB_TXT) as f:
      vocab = [line.rstrip('\n') for line in f]
    with open(SIZE_TXT) as f:
      content = [line.rstrip('\n') for line in f]
      if not content or len(content) != 1:
        raise ValueError("size.txt can only have one line")
    max_story_size, sentence_size = map(int, content[0].split(' '))

    global VOCAB
    VOCAB = vocab
    word_idx = dict((c, i + 1) for i, c in enumerate(vocab))
    vocab_size = len(word_idx) + 1  # +1 for nil word
    memory_size = min(FLAGS.memory_size, max_story_size)
    infer_data = load_infer_query(FLAGS.input_file)
    S, Q, A = get_kv_data.vectorize_data(infer_data, word_idx, sentence_size, memory_size)

    return get_kv_loss.MemN2N_KV(FLAGS.batch_size,
                                          vocab_size,
                                          sentence_size,
                                          sentence_size,
                                          memory_size,
                                          memory_size,
                                          FLAGS.embedding_size,
                                          S, Q, S, A,
                                          FLAGS.feature_size,
                                          FLAGS.hops,
                                          FLAGS.reader,
                                          FLAGS.l2_lambda,
                                          name='keyValueMemN2N')

def get_available_device():
    num_gpu = context.num_gpus()
    return '/device:GPU:0' if num_gpu else '/device:CPU:0'

def inference():
    """Inference from a single image file or an image file list,
       this is only supported for models under nets/.
       Only imagenet and customized datasets supported currently.

    Return:
        Predict target labels from previous train welled checkpoints
        with the full label files.
    """
    with tf.Graph().as_default():

      # Create session 
      gpu_options = tf.GPUOptions(per_process_gpu_memory_fraction=0.3)
      sess = tf.Session(config=tf.ConfigProto(gpu_options=gpu_options,
                                              allow_soft_placement=True,
                                              log_device_placement=True))

      infer_device = get_available_device()
      with tf.device(infer_device):
        memn2n = create_model() 

      with tf.device(infer_device), sess.as_default():
        # Restore variables from checkpoint file into graph
        saver = tf.train.Saver()
        ckpt = tf.train.latest_checkpoint(FLAGS.checkpoint_dir)
        load(saver, sess, ckpt)

        # Calculate probabilities with softmax activation
        prob_val = sess.run(memn2n.probs)

        answers = []
        if isinstance(prob_val, np.ndarray):
          index = np.argmax(prob_val, axis=1)
          for i, j in enumerate(index):
            print("Probabilities are: %s" % prob_val[i][j])
            print("Predict answer is: %s" % VOCAB[j-1])
            answers.append(VOCAB[j-1])
        else:
          raise ValueError("Unknown format probability result. ")

        _base = os.path.basename(FLAGS.input_file)
        _split = os.path.splitext(_base)
        if len(_split) == 2:
          output_file = _split[0] + '_predict' + _split[1]
        else:
          raise ValueError("Unknown inferfile format.")

        with open(FLAGS.output_dir + output_file, "w") as f:
          for a in answers:
            f.write(a + "\n")

def main(argv=None):  # pylint: disable=unused-argument
    pre_check()
    inference()

if __name__ == '__main__':
    tf.app.run()
